# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"create_markdown_table": "00_utils.ipynb",
         "set_up_logging": "00_utils.ipynb",
         "request_access_token": "01_eumetsat.ipynb",
         "query_data_products": "01_eumetsat.ipynb",
         "format_dt_str": "01_eumetsat.ipynb",
         "identify_available_datasets": "01_eumetsat.ipynb",
         "dataset_id_to_link": "01_eumetsat.ipynb",
         "json_extract": "01_eumetsat.ipynb",
         "extract_metadata": "01_eumetsat.ipynb",
         "metadata_maps": "01_eumetsat.ipynb",
         "check_valid_request": "01_eumetsat.ipynb",
         "DownloadManager": "01_eumetsat.ipynb",
         "get_dir_size": "01_eumetsat.ipynb",
         "debug_fp": "01_eumetsat.ipynb",
         "log": "01_eumetsat.ipynb",
         "get_filesize_megabytes": "01_eumetsat.ipynb",
         "eumetsat_filename_to_datetime": "01_eumetsat.ipynb",
         "compress_downloaded_files": "01_eumetsat.ipynb",
         "upload_compressed_files": "01_eumetsat.ipynb",
         "calculate_x_offset": "02_reproj.ipynb",
         "get_seviri_area_def": "02_reproj.ipynb",
         "load_scene": "02_reproj.ipynb",
         "construct_area_def": "02_reproj.ipynb",
         "construct_TM_area_def": "02_reproj.ipynb",
         "reproj_with_manual_grid": "02_reproj.ipynb",
         "reproj_to_xarray": "02_reproj.ipynb",
         "full_scene_pyresample": "02_reproj.ipynb",
         "full_scene_pyinterp": "02_reproj.ipynb",
         "Reprojector": "02_reproj.ipynb",
         "is_retriable": "03_zarr.ipynb",
         "load_from_zarr_bucket": "03_zarr.ipynb",
         "gcsfs.utils.is_retriable": "03_zarr.ipynb",
         "add_constant_coord_to_da": "03_zarr.ipynb",
         "Compressor": "03_zarr.ipynb",
         "save_da_to_zarr": "03_zarr.ipynb",
         "list_blobs_with_prefix": "04_gcp.ipynb",
         "blob_metadata": "04_gcp.ipynb",
         "upload_blob": "04_gcp.ipynb",
         "get_eumetsat_filenames": "04_gcp.ipynb",
         "write_metadata_to_gcp": "04_gcp.ipynb",
         "query": "04_gcp.ipynb",
         "sql_query": "04_gcp.ipynb",
         "project_id": "04_gcp.ipynb",
         "download_eumetsat_files": "05_pipeline.ipynb",
         "download_latest_eumetsat_files": "05_pipeline.ipynb",
         "df_metadata_to_dt_to_fp_map": "05_pipeline.ipynb",
         "reproject_datasets": "05_pipeline.ipynb",
         "compress_and_save_datasets": "05_pipeline.ipynb",
         "download_latest_data_pipeline": "05_pipeline.ipynb",
         "user_key": "05_pipeline.ipynb",
         "user_secret": "05_pipeline.ipynb",
         "slack_webhook_url": "05_pipeline.ipynb",
         "slack_id": "05_pipeline.ipynb"}

modules = ["utils.py",
           "eumetsat.py",
           "reproj.py",
           "io.py",
           "gcp_helpers.py",
           "mario.py"]

doc_url = "https://AyrtonB.github.io/satip/"

git_url = "https://github.com/AyrtonB/satip/tree/master/"

def custom_doc_links(name):
    return None




# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_pipeline.ipynb (unless otherwise specified).

__all__ = ['download_eumetsat_files', 'download_latest_eumetsat_files', 'df_metadata_to_dt_to_fp_map',
           'reproject_datasets', 'compress_and_save_datasets', 'download_latest_data_pipeline', 'user_key',
           'user_secret', 'slack_webhook_url', 'slack_id']

# Cell
# Imports
import pandas as pd
import xarray as xr

from satip import eumetsat, reproj, io, gcp_helpers
from dagster import execute_pipeline, pipeline, solid, Field

import os
import dotenv
dotenv.load_dotenv('.env')

# Loading environment vars
user_key = os.environ.get('USER_KEY')
user_secret = os.environ.get('USER_SECRET')
slack_webhook_url = os.environ.get('SLACK_WEBHOOK_URL')
slack_id = os.environ.get('SLACK_ID')

# Pipeline components
@solid(
    config_schema = {
        'user_key': Field(str, default_value=user_key, is_required=False),
        'user_secret': Field(str, default_value=user_secret, is_required=False),
        'slack_webhook_url': Field(str, default_value=slack_webhook_url, is_required=False),
        'slack_id': Field(str, default_value=slack_id, is_required=False)
    }
)
def download_eumetsat_files(context, start_date: str, end_date: str, data_dir: str, metadata_db_fp: str, debug_fp: str, table_id: str, project_id: str):
    dm = eumetsat.DownloadManager(context.solid_config['user_key'], context.solid_config['user_secret'], data_dir, metadata_db_fp, debug_fp, slack_webhook_url=context.solid_config['slack_webhook_url'], slack_id=context.solid_config['slack_id'])
    df_new_metadata = dm.download_datasets(start_date, end_date)

    if df_new_metadata is None:
        df_new_metadata = pd.DataFrame(columns=['result_time', 'file_name'])
    else:
        gcp_helpers.write_metadata_to_gcp(df_new_metadata, table_id, project_id, append=True)

    return df_new_metadata

# Pipeline components
@solid(
    config_schema = {
        'user_key': Field(str, default_value=user_key, is_required=False),
        'user_secret': Field(str, default_value=user_secret, is_required=False),
        'slack_webhook_url': Field(str, default_value=slack_webhook_url, is_required=False),
        'slack_id': Field(str, default_value=slack_id, is_required=False)
    }
)
def download_latest_eumetsat_files(context, data_dir: str, metadata_db_fp: str, debug_fp: str, table_id: str, project_id: str):
    sql_query = f'select * from {table_id} where result_time = (select max(result_time) from {table_id})'
    start_date = gcp_helpers.query(sql_query, project_id)['result_time'].iloc[0].strftime('%Y-%m-%d %H:%M')
    end_date = pd.Timestamp.now().strftime('%Y-%m-%d %H:%M')

    dm = eumetsat.DownloadManager(context.solid_config['user_key'], context.solid_config['user_secret'], data_dir, metadata_db_fp, debug_fp, slack_webhook_url=context.solid_config['slack_webhook_url'], slack_id=context.solid_config['slack_id'])
    df_new_metadata = dm.download_datasets(start_date, end_date)

    if df_new_metadata is None:
        df_new_metadata = pd.DataFrame(columns=['result_time', 'file_name'])
    else:
        gcp_helpers.write_metadata_to_gcp(df_new_metadata, table_id, project_id, append=True)

    return df_new_metadata

@solid()
def df_metadata_to_dt_to_fp_map(_, df_new_metadata, data_dir: str) -> dict:
    """
    Here we'll then identify downloaded files in
    the metadata dataframe and return a mapping
    between datetimes and filenames
    """

    datetime_to_filename = (df_new_metadata
                            .set_index('result_time')
                            ['file_name']
                            .drop_duplicates()
                            .to_dict()
                           )

    datetime_to_filepath = {
        datetime: f"{data_dir}/{filename}.nat"
        for datetime, filename
        in datetime_to_filename.items()
        if filename != {}
    }

    return datetime_to_filepath

@solid()
def reproject_datasets(_, datetime_to_filepath: dict, new_coords_fp: str, new_grid_fp: str):
    reprojector = reproj.Reprojector(new_coords_fp, new_grid_fp)

    reprojected_dss = [
        (reprojector
         .reproject(filepath, reproj_library='pyinterp')
         .pipe(io.add_constant_coord_to_da, 'time', pd.to_datetime(datetime))
        )
        for datetime, filepath
        in datetime_to_filepath.items()
    ]

    if len(reprojected_dss) > 0:
        ds_combined_reproj = xr.concat(reprojected_dss, 'time', coords='all', data_vars='all')
        return ds_combined_reproj
    else:
        return xr.Dataset()

@solid()
def compress_and_save_datasets(_, ds_combined_reproj, zarr_bucket: str, var_name: str='stacked_eumetsat_data'):
    # Handle case where no new data exists
    if len(ds_combined_reproj.dims) == 0:
        return

    # Compressing the datasets
    compressor = io.Compressor()

    var_name = var_name
    da_compressed = compressor.compress(ds_combined_reproj[var_name])

    # Saving to Zarr
    io.save_da_to_zarr(da_compressed, zarr_bucket)

    return

# Pipeline
@pipeline
def download_latest_data_pipeline():
    # Retrieving data, reprojecting, compressing, and saving to GCP
    df_new_metadata = download_latest_eumetsat_files()
    datetime_to_filepath = df_metadata_to_dt_to_fp_map(df_new_metadata)
    ds_combined_reproj = reproject_datasets(datetime_to_filepath)
    compress_and_save_datasets(ds_combined_reproj)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/03_zarr.ipynb (unless otherwise specified).

__all__ = ['is_retriable', 'load_from_zarr_bucket', 'add_constant_coord_to_da', 'Compressor', 'save_da_to_zarr']

# Cell
import pandas as pd
import xarray as xr
import numpy as np
import dask
import zarr
import gcsfs
import numcodecs

# Cell
def is_retriable(exception):
    """Returns True if this exception is retriable."""
    errors = list(range(500, 505)) + [
        400, # Jack's addition.  Google Cloud occasionally throws Bad Requests for no apparent reason.
        408,# Request Timeout
        429, # Too Many Requests
    ]

    errors += [str(e) for e in errors]

    if isinstance(exception, gcsfs.utils.HttpError):
        return exception.code in errors

    return isinstance(exception, gcsfs.utils.RETRIABLE_EXCEPTIONS)

gcsfs.utils.is_retriable = is_retriable

def load_from_zarr_bucket(zarr_bucket):
    gcs = gcsfs.GCSFileSystem()
    store = gcsfs.GCSMap(root=zarr_bucket, gcs=gcs)
    ds = xr.open_zarr(store, consolidated=True)

    return ds

# Cell
def add_constant_coord_to_da(da, coord_name, coord_val):
    """
    Adds a new coordinate with a
    constant value to the DataArray

    Parameters
    ----------
    da : xr.DataArray
        DataArrray which will have the new coords added to it
    coord_name : str
        Name for the new coordinate dimensions
    coord_val
        Value that will be assigned to the new coordinates

    Returns
    -------
    da : xr.DataArray
        DataArrray with the new coords added to it

    """

    da = (da
          .assign_coords({coord_name:coord_val})
          .expand_dims(coord_name)
         )

    return da

class Compressor:
    def __init__(self,
                 bits_per_pixel=10,
                 mins=np.array([-1.2278595, -2.5118103, -64.83977, 63.404694, 2.844452, 199.10002, -17.254883, -26.29155, -1.1009827, -2.4184198, 199.57048, 198.95093]),
                 maxs=np.array([103.90016, 69.60857, 339.15588, 340.26526, 317.86752, 313.2767, 315.99194, 274.82297, 93.786545, 101.34922, 249.91806, 286.96323]),
                 variable_order=['HRV', 'IR_016', 'IR_039', 'IR_087', 'IR_097', 'IR_108', 'IR_120', 'IR_134', 'VIS006', 'VIS008', 'WV_062', 'WV_073']
                ):

        locals_ = locals()
        attrs_to_add = ['bits_per_pixel', 'mins', 'maxs', 'variable_order']

        for attr in attrs_to_add:
            setattr(self, attr, locals_[attr])

        return

    def fit(self, da, dims=['time', 'y', 'x']):
        self.mins = da.min(dims).compute()
        self.maxs = da.max(dims).compute()
        self.variable_order = da.coords['variable'].values

        print(f'The mins are: {self.mins}')
        print(f'The maxs are: {self.maxs}')
        print(f'The variable order is: {self.variable_order}')

        return

    def compress(self, da):
        da_meta = da.attrs

        for attr in ['mins', 'maxs']:
            assert getattr(self, attr) is not None, f'{attr} must be set in initialisation or through `fit`'

        if 'time' not in da.dims:
            time = pd.to_datetime(da_meta['end_time'])
            da = add_constant_coord_to_da(da, 'time', time)

        da = (da
              .reindex({'variable': self.variable_order})
              .transpose('time', 'y', 'x', 'variable')
             )

        upper_bound = (2 ** self.bits_per_pixel) - 1
        new_max = self.maxs - self.mins

        da -= self.mins
        da /= new_max
        da *= upper_bound

        da = (da
              .fillna(-1)
              .round()
              .astype(np.int16)
             )

        da.attrs = {'meta': str(da_meta)} # Must be serialisable

        return da

# Cell
def save_da_to_zarr(da, zarr_bucket, dim_order=['time', 'x', 'y', 'variable'], zarr_mode='a'):
    da = da.transpose(*dim_order)
    _, y_size, x_size, _ = da.shape
    out_store = gcsfs.GCSMap(root=zarr_bucket, gcs=gcsfs.GCSFileSystem())

    chunks = (36, y_size, x_size, 1)

    ds = xr.Dataset({'stacked_eumetsat_data': da.chunk(chunks)})

    zarr_mode_to_extra_kwargs = {
        'a': {
            'append_dim': 'time'
        },
        'w': {
            'encoding': {
                'stacked_eumetsat_data': {
                    'compressor': numcodecs.Blosc(cname='zstd', clevel=5),
                    'chunks': chunks
                }
            }
        }
    }

    assert zarr_mode in ['a', 'w'], '`zarr_mode` must be one of: `a`, `w`'
    extra_kwargs = zarr_mode_to_extra_kwargs[zarr_mode]

    ds.to_zarr(out_store, mode=zarr_mode, consolidated=True, **extra_kwargs)

    return